from pydantic import BaseModel
from pydantic import BaseModel, Field, schema, validator
from typing import List

class PaginationResponseModel(BaseModel):
    total_rows: int
    page: int
    page_size: int
    is_first_page: bool
    is_last_page: bool


class ExternalCategoryModel(BaseModel):
    external_id: str
    name: str
    parent_category: str
    preview_url: str


class CategoryModel(ExternalCategoryModel):
    name: str
    parent_category: str = ''
    preview_url: str = ''

    @validator('name')
    def not_empty_string(cls, value):
        if not value:
            raise ValueError('Name cannot be an empty string.')
        return value
    

class CategoryResponseModel(BaseModel):
    id: str
    name: str
    parent_category: str
    preview_url: str

    # TODO: add bitrix category id here
    external_id: str = None

class CategoryListResponseModel(BaseModel):
    categories: List[CategoryResponseModel]
    page_info: PaginationResponseModel


NOCODB_CATEGORIES="Categories"
NOCODB_PRODUCTS="Products"
NOCODB_STATUSES="Order statuses"
NOCODB_BOT_MESSAGES="Bot messages"
NOCODB_ORDERS="Orders"

# NOCODB_CATEGORIES="Категории"
# NOCODB_PRODUCTS="Товары"
# NOCODB_STATUSES="Статусы заказов"
# NOCODB_BOT_MESSAGES="Сообщения бота"
# NOCODB_ORDERS="Заказы"


from pydantic import BaseModel
from typing import Any, List,Dict,Optional
from datetime import datetime

from pydantic import BaseModel, Field, schema, validator

class ExtraAttribute(BaseModel):
    name: str
    description: Optional[str]

class ExternalProductModel(BaseModel):
    external_id: str
    category: Optional[List[str]]
    category_name: Optional[List[str]]
    name: str
    description: Optional[str]
    price: Optional[float]
    final_price: Optional[float]
    currency: Optional[str]
    stock_qty: int
    orders_qty: int = Field(0, hidden_field=True)
    created: datetime = Field(default=datetime.now, hidden_field=True)
    updated: datetime = Field(default=datetime.now, hidden_field=True)
    preview_url: List[str] = []
    extra_attributes: List[ExtraAttribute] = []

class ProductModel(BaseModel):
    id: Optional[str]
    external_id: Optional[str]
    product_properties: Optional[List[str]]
    categories_structure: Optional[Dict[str, Any]]
    name: str
    description: Optional[str]
    price: Optional[float]
    final_price: Optional[float]
    currency: Optional[str]
    stock_qty: int
    orders_qty: int = Field(0, hidden_field=True)
    created: datetime = Field(default=datetime.now, hidden_field=True)
    updated: datetime = Field(default=datetime.now, hidden_field=True)
    preview_url: List[str] = []
    extra_attributes: List[ExtraAttribute] = []



from typing import List, Optional, Dict, Union
import httpx
import requests
import io
from loguru import logger
from tgshops_integrations.nocodb_connector.model_mapping import ID_FIELD


def custom_key_builder(func, *args, **kwargs) -> str:
    """
    Custom key builder for caching.
    Excludes 'self' by starting args processing from args[1:].
    """
    args_key_part = "-".join(str(arg) for arg in args[1:])
    kwargs_key_part = "-".join(f"{key}-{value}" for key, value in sorted(kwargs.items()))
    return f"{func.__name__}-{args_key_part}-{kwargs_key_part}"


class NocodbClient:
    def __init__(self, NOCODB_HOST: Optional[str] = None, NOCODB_API_KEY: Optional[str] = None, SOURCE: Optional[str] = None):
        self.NOCODB_HOST = NOCODB_HOST
        self.NOCODB_API_KEY = NOCODB_API_KEY
        self.SOURCE = SOURCE
        self.httpx_client = httpx.AsyncClient(timeout=60.0)
        self.httpx_client.headers = {"xc-token": self.NOCODB_API_KEY}

    def construct_get_params(
        self,
        required_fields: Optional[List[str]] = None,
        projection: Optional[List[str]] = None,
        extra_where: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None
    ) -> Dict[str, Union[str, int]]:
        """
        Constructs GET parameters for API requests.
        """
        params = {}
        if projection:
            params["fields"] = ','.join(projection)
        if required_fields:
            params["where"] = "~and".join(f"({field},isnot,null)" for field in required_fields)
        if extra_where:
            params["where"] = f"{params.get('where', '')}~and{extra_where}" if params.get("where") else extra_where
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        return params

    async def get_table_records(
        self,
        table_name: str,
        required_fields: Optional[List[str]] = None,
        projection: Optional[List[str]] = None,
        extra_where: Optional[str] = None,
        limit: Optional[int] = None
    ) -> List[dict]:
        """
        Fetches records from a specified table.
        """
        url = f"{self.NOCODB_HOST}/tables/{table_name}/records"
        params = self.construct_get_params(required_fields, projection, extra_where, limit=limit)
        response = await self.httpx_client.get(url, params=params)
        if response.status_code == 200:
            return response.json().get("list", [])
        logger.error(f"Error fetching records: {response.text}")
        raise Exception(response.text)
    
    async def get_table_records_v2(self,
                                   table_name: str,
                                   required_fields: list = None,
                                   projection: list = None,
                                   extra_where: str = None,
                                   offset: int = None,
                                   limit: int = 25) -> dict:
        url = f"{self.NOCODB_HOST}/tables/{table_name}/records"
        extra_params = self.construct_get_params(required_fields, projection, extra_where, offset=offset, limit=limit)
        response = await self.httpx_client.get(url, params=extra_params)
        if response.status_code == 200:
            return response.json()
        raise Exception(response.text)

    async def get_table_record(self,
                               table_name: str,
                               record_id: str,
                               required_fields: list = None,
                               projection: list = None) -> dict:
        url = f"{self.NOCODB_HOST}/tables/{table_name}/records/{record_id}"
        extra_params = self.construct_get_params(required_fields, projection)
        response = await self.httpx_client.get(url, params=extra_params)
        if response.status_code == 200:
            return response.json()
        raise Exception(response.text)

    async def create_table_record(self, table_name: str, record: dict) -> dict:
        """
        Creates a new record in a table.
        """
        url = f"{self.NOCODB_HOST}/tables/{table_name}/records"
        response = await self.httpx_client.post(url, json=record)
        if response.status_code == 200:
            record["id"] = response.json().get("id") or response.json().get("Id")
            return record
        logger.error(f"Error creating record: {response.text}")
        raise Exception(response.text)

    async def delete_table_record(self, table_name: str, record_id: str) -> dict:
        """
        Deletes a record from a specified table.
        """
        url = f"{self.NOCODB_HOST}/tables/{table_name}/records"
        response = requests.delete(url, json={"Id": record_id}, headers=self.httpx_client.headers)
        if response.status_code == 200:
            logger.info(f"Deleted record {record_id}")
            return response.json()
        logger.error(f"Error deleting record: {response.text}")
        raise Exception(response.text)

    async def get_product_categories(self, table_id: str, table_name: str) -> Dict[str, str]:
        """
        Fetches product categories from a specified table.
        """
        url = f"{self.NOCODB_HOST}/tables/{table_id}/records"
        params = self.construct_get_params(limit=75)
        response = await self.httpx_client.get(url, params=params)
        if response.status_code == 200:
            return {category[table_name]: category["Id"] for category in response.json().get("list", [])}
        logger.error(f"Error fetching categories: {response.text}")
        raise Exception(response.text)

    async def create_product_category(
        self, table_id: str, category_name: str, table_name: str, category_id: int = 0
    ) -> dict:
        """
        Creates a new product category in a specified table.
        """
        url = f"{self.NOCODB_HOST}/tables/{table_id}/records"
        record = {table_name: category_name, "Id": category_id}
        response = await self.httpx_client.post(url, json=record)
        if response.status_code == 200:
            return record
        logger.error(f"Error creating product category: {response.text}")
        raise Exception(response.text)

    async def get_table_meta(self, table_name: str) -> dict:
        """
        Fetches metadata of a table.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/meta/tables/{table_name}"
        response = await self.httpx_client.get(url)
        if response.status_code == 200:
            return response.json()
        logger.error(f"Error fetching table metadata: {response.text}")
        raise Exception(response.text)

    async def get_all_tables(self, source: Optional[str] = None) -> Dict[str, str]:
        """
        Fetches all tables from the specified project source.
        """
        source = source or self.SOURCE
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/meta/projects/{source}/tables?includeM2M=false"
        response = await self.httpx_client.get(url)
        if response.status_code == 200:
            tables_info = response.json().get('list', [])
            self.tables_list = {table["title"]: table["id"] for table in tables_info}
            return self.tables_list
        logger.error(f"Failed to fetch tables: {response.text}")
        raise Exception(response.text)

    async def get_sources(self) -> list:
        """
        Fetches all project sources.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/meta/projects/"
        response = await self.httpx_client.get(url)
        if response.status_code == 200:
            return response.json().get('list', [])
        logger.error(f"Failed to fetch sources: {response.text}")
        raise Exception(response.text)

    async def get_table_meta(self, table_name: str) -> dict:
        """
        Fetches metadata of a specified table.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/meta/tables/{table_name}"
        response = await self.httpx_client.get(url)
        if response.status_code == 200:
            return response.json()
        logger.error(f"Failed to fetch table metadata: {response.text}")
        raise Exception(response.text)

    async def create_table_column(self, table_name: str, name: str) -> dict:
        """
        Creates a new column in the specified table.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/meta/tables/{table_name}/columns"
        payload = {
            "column_name": name,
            "dt": "character varying",
            "dtx": "specificType",
            "ct": "varchar(45)",
            "clen": 45,
            "dtxp": "45",
            "dtxs": "",
            "altered": 1,
            "uidt": "SingleLineText",
            "uip": "",
            "uicn": "",
            "title": name
        }
        response = await self.httpx_client.post(url, json=payload)
        if response.status_code == 200:
            return response.json()
        logger.error(f"Failed to create table column: {response.text}")
        raise Exception(response.text)

    async def link_table_record(
        self,
        base_id: str,
        fk_model_id: str,
        record_id: str,
        source_column_id: str,
        linked_record_id: str
    ) -> dict:
        """
        Links a record to another record in a many-to-many relationship.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/data/noco/{base_id}/{fk_model_id}/{record_id}/mm/{source_column_id}/{linked_record_id}"
        response = await self.httpx_client.post(url, headers=self.httpx_client.headers)
        if response.status_code == 200:
            return response.json()
        logger.error(f"Failed to link table record: {response.text}")
        raise Exception(response.text)

    async def unlink_table_record(
        self,
        base_id: str,
        fk_model_id: str,
        record_id: str,
        source_column_id: str,
        linked_record_id: str
    ) -> dict:
        """
        Unlinks a record from another record in a many-to-many relationship.
        """
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/data/noco/{base_id}/{fk_model_id}/{record_id}/mm/{source_column_id}/{linked_record_id}"
        response = await self.httpx_client.delete(url)
        if response.status_code == 200:
            return response.json()
        logger.error(f"Failed to unlink table record: {response.text}")
        raise Exception(response.text)

    async def save_image_to_nocodb(
        self,
        image_url: str,
        image_name: str,
        source_column_id: str,
        product_table_name: str,
        images_column_id: str
    ) -> Optional[str]:
        """
        Saves an image to NocoDB's storage.
        """
        try:
            response = requests.get(image_url)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Error loading image from URL {image_url}: {e}")
            return None

        file = io.BytesIO(response.content)
        file_size = file.getbuffer().nbytes

        if not file_size:
            logger.error(f"Image file from {image_url} is empty.")
            return None

        files = {'file': (image_name, file, 'image/jpeg')}
        url = f"{self.NOCODB_HOST.replace('/api/v2', '/api/v1')}/db/storage/upload?path=noco/{source_column_id}/{product_table_name}/{images_column_id}"
        try:
            response = await self.httpx_client.post(url, files=files, timeout=httpx.Timeout(200.0))
            if response.status_code == 200:
                return response.json()[0].get('url', None)
            logger.error(f"Error saving image {image_name}: {response.text}")
        except Exception as e:
            logger.error(f"Unexpected error saving image {image_name}: {e}")
        return None



import json
import secrets
import importlib.util
from pathlib import Path

from tgshops_integrations.models.categories import CategoryModel, CategoryResponseModel, PaginationResponseModel
from tgshops_integrations.models.products import ExtraAttribute, ProductModel


# Helper function to load `config.py` dynamically
def load_config(config_path: str):
    """
    Dynamically load a config file from the provided path.
    """
    config_path = Path(config_path)
    if config_path.exists():
        spec = importlib.util.spec_from_file_location("config", config_path)
        config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config)
        return config
    else:
        raise FileNotFoundError(f"Configuration file not found at {config_path}")


# Initialize model mapping constants dynamically from config
def initialize_model_mapping(config_path: str):
    """
    Load and initialize global constants for model mapping from a config file.
    """
    global CATEGORY_IMAGE_FIELD, ID_FIELD, CATEGORY_NAME_FIELD, CATEGORY_PARENT_ID_FIELD, CATEGORY_PARENT_FIELD
    global CATEGORY_ID_OF_CATEGORY_FIELD, PRODUCT_NAME_FIELD, PRODUCT_DESCRIPTION_FIELD, PRODUCT_PRICE_FIELD
    global PRODUCT_CURRENCY_FIELD, PRODUCT_STOCK_FIELD, PRODUCT_CATEGORY_NAME_FIELD, PRODUCT_CATEGORY_ID_FIELD
    global PRODUCT_IMAGE_FIELD, PRODUCT_DISCOUNT_PRICE_FIELD, PRODUCT_CATEGORY_ID_LOOKUP_FIELD, PRODUCT_IMAGES_LOOKUP_FIELD
    global PRODUCT_REQUIRED_OPTIONS_FIELD, PRODUCT_CATEGORIES_EXTRA_OPTIONS_FIELD, PRODUCT_CATEGORIES_EXTRA_OPTION_NAMES_FIELD
    global PRODUCT_EXTRA_CHOICE_REQUIRED_FIELD, PRODUCT_ID_FIELD, PRODUCT_EXTERNAL_ID, PRODUCT_CHECKOUT_MODE
    global NEW_ID_FIELD, NOCODB_CHECKOUT_MODES,PRODUCT_CATEGORY_STRUCTURE

    config = load_config(config_path)

    CATEGORY_IMAGE_FIELD = config.CATEGORY_IMAGE_FIELD
    ID_FIELD = config.ID_FIELD
    CATEGORY_NAME_FIELD = config.CATEGORY_NAME_FIELD
    CATEGORY_PARENT_ID_FIELD = config.CATEGORY_PARENT_ID_FIELD
    CATEGORY_PARENT_FIELD = config.CATEGORY_PARENT_FIELD
    CATEGORY_ID_OF_CATEGORY_FIELD = config.CATEGORY_ID_OF_CATEGORY_FIELD

    PRODUCT_NAME_FIELD = config.PRODUCT_NAME_FIELD
    PRODUCT_DESCRIPTION_FIELD = config.PRODUCT_DESCRIPTION_FIELD
    PRODUCT_PRICE_FIELD = config.PRODUCT_PRICE_FIELD
    PRODUCT_CURRENCY_FIELD = config.PRODUCT_CURRENCY_FIELD
    PRODUCT_STOCK_FIELD = config.PRODUCT_STOCK_FIELD
    PRODUCT_CATEGORY_NAME_FIELD = config.PRODUCT_CATEGORY_NAME_FIELD
    PRODUCT_CATEGORY_STRUCTURE=config.PRODUCT_CATEGORY_STRUCTURE
    PRODUCT_CATEGORY_ID_FIELD = config.PRODUCT_CATEGORY_ID_FIELD
    PRODUCT_IMAGE_FIELD = config.PRODUCT_IMAGE_FIELD
    PRODUCT_DISCOUNT_PRICE_FIELD = config.PRODUCT_DISCOUNT_PRICE_FIELD
    PRODUCT_CATEGORY_ID_LOOKUP_FIELD = config.PRODUCT_CATEGORY_ID_LOOKUP_FIELD
    PRODUCT_IMAGES_LOOKUP_FIELD = config.PRODUCT_IMAGES_LOOKUP_FIELD
    PRODUCT_REQUIRED_OPTIONS_FIELD = config.PRODUCT_REQUIRED_OPTIONS_FIELD
    PRODUCT_CATEGORIES_EXTRA_OPTIONS_FIELD = config.PRODUCT_CATEGORIES_EXTRA_OPTIONS_FIELD
    PRODUCT_CATEGORIES_EXTRA_OPTION_NAMES_FIELD = config.PRODUCT_CATEGORIES_EXTRA_OPTION_NAMES_FIELD
    PRODUCT_EXTRA_CHOICE_REQUIRED_FIELD = config.PRODUCT_EXTRA_CHOICE_REQUIRED_FIELD
    PRODUCT_ID_FIELD = config.PRODUCT_ID_FIELD
    PRODUCT_EXTERNAL_ID = config.PRODUCT_EXTERNAL_ID
    PRODUCT_CHECKOUT_MODE = config.PRODUCT_CHECKOUT_MODE
    NEW_ID_FIELD = config.NEW_ID_FIELD
    NOCODB_CHECKOUT_MODES = config.NOCODB_CHECKOUT_MODES


def get_pagination_info(page_info: dict) -> PaginationResponseModel:
    """
    Parses pagination information into a model.
    """
    return PaginationResponseModel(
        total_rows=page_info['totalRows'],
        page=page_info['page'],
        page_size=page_info['pageSize'],
        is_first_page=page_info['isFirstPage'],
        is_last_page=page_info['isLastPage']
    )


def parse_category_data(data: dict) -> CategoryResponseModel:
    """
    Parses raw category data into a structured model.
    """
    preview_url = data.get(CATEGORY_IMAGE_FIELD, [{}])[0].get("url", "") if data.get(CATEGORY_IMAGE_FIELD) else ""
    return CategoryResponseModel(
        id=str(data[ID_FIELD]),
        name=data.get(CATEGORY_NAME_FIELD, ""),
        parent_category=str(data.get(CATEGORY_PARENT_ID_FIELD, 0)),
        preview_url=preview_url,
    )


def dump_category_data(data: CategoryModel) -> dict:
    """
    Converts a CategoryModel into a dictionary suitable for API calls.
    """
    return {
        CATEGORY_NAME_FIELD: data.name,
        CATEGORY_PARENT_FIELD: data.parent_category,
        CATEGORY_IMAGE_FIELD: [
            {"url": data.preview_url, "title": f"{secrets.token_hex(6)}.jpeg", "mimetype": "image/jpeg"}
        ]
    }


def dump_product_data(data: ProductModel) -> dict:
    """
    Converts a ProductModel into a dictionary suitable for API calls.
    """
    preview_url = [
        {"url": image_url, "title": f"{secrets.token_hex(6)}.jpeg", "mimetype": "image/jpeg"}
        for image_url in data.preview_url
    ] if data.preview_url else []

    return {
        PRODUCT_NAME_FIELD: data.name,
        PRODUCT_DESCRIPTION_FIELD: data.description,
        PRODUCT_PRICE_FIELD: data.price,
        PRODUCT_CURRENCY_FIELD: data.currency,
        PRODUCT_STOCK_FIELD: data.stock_qty,
        PRODUCT_CATEGORY_NAME_FIELD: [data.category_name] if data.category_name else None,
        PRODUCT_CATEGORY_ID_FIELD: [{"Id": data.category}] if data.category else None,
        PRODUCT_IMAGE_FIELD: preview_url,
        PRODUCT_DISCOUNT_PRICE_FIELD: data.final_price
    }


def dump_product_data_with_check(data: ProductModel, data_check: dict) -> dict:
    """
    Converts a ProductModel into a dictionary and validates categories.
    """
    preview_url = [
        {"url": image_url, "title": f"{secrets.token_hex(6)}.jpeg", "mimetype": "image/jpeg"}
        for image_url in data.preview_url
    ] if data.preview_url else []

    extra_data = {attr.name: attr.description for attr in data.extra_attributes}

    product_data = {
        PRODUCT_ID_FIELD: data.id,
        PRODUCT_EXTERNAL_ID: data.external_id,
        PRODUCT_NAME_FIELD: data.name,
        PRODUCT_DESCRIPTION_FIELD: data.description,
        PRODUCT_PRICE_FIELD: data.price,
        PRODUCT_CURRENCY_FIELD: data.currency,
        PRODUCT_STOCK_FIELD: data.stock_qty,
        PRODUCT_CATEGORY_NAME_FIELD: [data.product_properties] if data.product_properties else None,
        PRODUCT_CATEGORY_ID_FIELD: [{"Id": data_check[item]} for item in data.product_properties] if data.product_properties else None,
        PRODUCT_IMAGE_FIELD: preview_url,
        PRODUCT_CHECKOUT_MODE: NOCODB_CHECKOUT_MODES,
        PRODUCT_DISCOUNT_PRICE_FIELD: data.final_price,
    }

    if extra_data:
        product_data.update(extra_data)
    return product_data


async def parse_product_data(data: dict) -> ProductModel:
    """
    Parses raw product data into a ProductModel.
    """
    preview_url = [image['url'] for image in data.get(PRODUCT_IMAGE_FIELD, [])]

    # Dynamically add extra attributes
    extra_attributes = [
        ExtraAttribute(name=key, description=str(value))
        for key, value in data.items()
        if key not in {
            ID_FIELD, PRODUCT_NAME_FIELD, PRODUCT_DESCRIPTION_FIELD, PRODUCT_PRICE_FIELD, PRODUCT_CURRENCY_FIELD,
            PRODUCT_STOCK_FIELD, PRODUCT_CATEGORY_ID_FIELD, PRODUCT_IMAGE_FIELD, PRODUCT_CATEGORY_NAME_FIELD,
            PRODUCT_DISCOUNT_PRICE_FIELD, PRODUCT_CATEGORY_ID_LOOKUP_FIELD, PRODUCT_REQUIRED_OPTIONS_FIELD,
            PRODUCT_CATEGORIES_EXTRA_OPTIONS_FIELD, PRODUCT_CATEGORIES_EXTRA_OPTION_NAMES_FIELD,
            PRODUCT_EXTRA_CHOICE_REQUIRED_FIELD, "UpdatedAt", "CreatedAt"
        } and value is not None and isinstance(value, (str, int, float))
    ]

    product = ProductModel(
        id=str(data.get(ID_FIELD, data.get(NEW_ID_FIELD, ""))),
        external_id=data.get(PRODUCT_EXTERNAL_ID, ""),
        name=data.get(PRODUCT_NAME_FIELD, ""),
        description=data.get(PRODUCT_DESCRIPTION_FIELD, ""),
        price=data.get(PRODUCT_PRICE_FIELD, 0.0),
        final_price=data.get(PRODUCT_DISCOUNT_PRICE_FIELD, 0.0),
        currency=data.get(PRODUCT_CURRENCY_FIELD, "RUB"),
        stock_qty=data.get(PRODUCT_STOCK_FIELD, 0),
        preview_url=preview_url,
        product_properties=data.get(PRODUCT_CATEGORY_NAME_FIELD, []),
        categories_structure=data.get(PRODUCT_CATEGORY_STRUCTURE, {}),
        category=data.get(PRODUCT_CATEGORY_ID_LOOKUP_FIELD, []),
        extra_attributes=extra_attributes,
        extra_option_choice_required=bool(data.get(PRODUCT_EXTRA_CHOICE_REQUIRED_FIELD, [])),

        metadata=data
    )
    
    return product



from typing import List
from loguru import logger
from aiocache import cached
from tgshops_integrations.models.categories import CategoryModel, CategoryResponseModel, CategoryListResponseModel
from tgshops_integrations.models.products import ProductModel
from tgshops_integrations.nocodb_connector.client import custom_key_builder, NocodbClient
from tgshops_integrations.nocodb_connector.model_mapping import (
    # CATEGORY_IMAGE_FIELD,
    CATEGORY_NAME_FIELD,
    CATEGORY_PARENT_FIELD,
    CATEGORY_PARENT_ID_FIELD,
    PRODUCT_NAME_FIELD,
    CATEGORY_ID_OF_CATEGORY_FIELD,
    dump_category_data,
    get_pagination_info,
    parse_category_data,
)

class CategoryManager(NocodbClient):
    def __init__(
        self,
        table_id=None,
        logging=False,
        config_type=None,
        NOCODB_HOST=None,
        NOCODB_API_KEY=None,
        SOURCE=None,
        filter_buttons=None,
    ):
        super().__init__(NOCODB_HOST=NOCODB_HOST, NOCODB_API_KEY=NOCODB_API_KEY, SOURCE=SOURCE)
        self.NOCODB_HOST = NOCODB_HOST
        self.NOCODB_API_KEY = NOCODB_API_KEY
        self.SOURCE = SOURCE
        self.CONFIG_TYPE = config_type
        self.categories_table = table_id
        self.external_categories = {}
        self.logging = logging
        self.filter_categories = []
        self.filter_buttons = filter_buttons or []
        self.required_fields = [CATEGORY_NAME_FIELD]
        self.projection = ["Id", CATEGORY_NAME_FIELD, CATEGORY_PARENT_ID_FIELD, CATEGORY_ID_OF_CATEGORY_FIELD]

    @cached(ttl=30, key_builder=custom_key_builder)
    async def get_categories(self, table_id: str) -> List[CategoryModel]:
        records = await self.get_table_records(table_id, self.required_fields, self.projection)
        return [parse_category_data(record) for record in records]

    @cached(ttl=30, key_builder=custom_key_builder)
    async def get_categories_v2(self, table_id: str, offset: int = None, limit: int = None) -> CategoryListResponseModel:
        response = await self.get_table_records_v2(
            table_name=self.categories_table,
            required_fields=self.required_fields,
            projection=self.projection,
            offset=offset,
            limit=limit,
        )
        page_info = get_pagination_info(page_info=response['pageInfo'])
        categories = [parse_category_data(record) for record in response['list']]
        return CategoryListResponseModel(categories=categories, page_info=page_info)

    @cached(ttl=30, key_builder=custom_key_builder)
    async def get_category(self, table_id: str, category_id: str) -> CategoryModel:
        record = await self.get_table_record(self.categories_table, category_id, self.required_fields, self.projection)
        return parse_category_data(record)

    async def create_category(self, table_id: str, category: CategoryModel) -> CategoryModel:
        category_json = dump_category_data(category)
        record = await self.create_table_record(self.categories_table, category_json)
        return parse_category_data(record)

    @cached(ttl=30, key_builder=custom_key_builder)
    async def get_categories_in_category(self, table_id: str, category_id: str) -> List[CategoryModel]:
        extra_where = (
            f"({CATEGORY_PARENT_ID_FIELD},eq,{category_id})"
            if category_id
            else f"({CATEGORY_PARENT_FIELD},eq,0)"
        )
        records = await self.get_table_records(
            table_name=self.categories_table,
            required_fields=self.required_fields,
            projection=self.projection,
            extra_where=extra_where,
        )
        return [parse_category_data(record) for record in records]

    async def update_categories(self, external_products: List[ProductModel],with_properties: bool = False) -> None:
        self.categories = await self.get_product_categories(table_id=self.categories_table, table_name=PRODUCT_NAME_FIELD)
        categories_list = list(self.categories.keys())
        properties_to_create = []
        parent_id=0

        if with_properties:
            for product in external_products:
                for num, product_property in enumerate(product.product_properties[:len(self.filter_buttons)]):
                    if product_property not in categories_list:
                        # TODO: Needs restructuring
                        # parent_id = (
                        #     self.categories[self.filter_buttons[num]]
                        #     if self.filter_buttons
                        #     else self.categories[product.category_I_name[0]]
                        # )
                        
                        properties_to_create.append([product_property, parent_id])
                        categories_list.append(product_property)

        else:
            for product in external_products:
                # for item in product.categories_structure.values():
                #     if item not in categories_list:
                for product_property in product.product_properties:
                        if product_property not in categories_list:
                            properties_to_create.append([product_property, parent_id])
                            categories_list.append(product_property)

        if properties_to_create:
            properties_to_create.sort(key=lambda x: x[0])
            new_id = max(self.categories.values(), default=0) + 1

            for product_property, parent_id in properties_to_create:
                new_property = await self.create_product_category(
                    table_id=self.categories_table,
                    category_name=product_property,
                    category_id=new_id,
                    table_name=PRODUCT_NAME_FIELD,
                )
                if self.logging:
                    logger.info(f"New Category: {new_property}")
                new_id += 1
                
        self.categories = await self.get_product_categories(table_id=self.categories_table, table_name=PRODUCT_NAME_FIELD)

    async def link_categories(self, parent_id: int, child_id: int):
        metadata = await self.get_table_meta(self.categories_table)
        linked_column = next((col for col in metadata['columns'] if col["title"] == "Set parent category" and col["uidt"] == "Links"), None)

        if linked_column:
            await self.link_table_record(
                linked_column["base_id"],
                linked_column["fk_model_id"],
                child_id,
                linked_column["id"],
                parent_id)

    async def unlink_categories(self, parent_id: int, child_id: int):
        metadata = await self.get_table_meta(self.categories_table)
        linked_column = next((col for col in metadata['columns'] if col["uidt"] == "Links"), None)

        if linked_column:
            await self.unlink_table_record(
                linked_column["base_id"],
                linked_column["fk_model_id"],
                parent_id,
                linked_column["id"],
                child_id,
            )

    async def map_categories(self, external_products: List[ProductModel]) -> List[ProductModel]:
        for num, product in enumerate(external_products):
            if not product.category and product.product_properties:
                external_products[num].category = [
                    str(self.categories[property_name]) for property_name in product.product_properties
                ]
        return external_products



from typing import List, Optional
import hashlib
from aiocache import cached
from loguru import logger
from tgshops_integrations.models.products import ProductModel
from tgshops_integrations.nocodb_connector.client import custom_key_builder, NocodbClient
from tgshops_integrations.nocodb_connector.model_mapping import (
    dump_product_data,
    dump_product_data_with_check,
    get_pagination_info,
    parse_product_data,
    ID_FIELD,
    PRODUCT_CATEGORY_ID_LOOKUP_FIELD,
    PRODUCT_NAME_FIELD,
    PRODUCT_PRICE_FIELD,
    PRODUCT_STOCK_FIELD,
    PRODUCT_EXTERNAL_ID,
    PRODUCT_IMAGES_LOOKUP_FIELD,
)

class ProductManager(NocodbClient):
    def __init__(self, table_id=None, logging=False, NOCODB_HOST=None, NOCODB_API_KEY=None, SOURCE=None):
        super().__init__(NOCODB_HOST=NOCODB_HOST, NOCODB_API_KEY=NOCODB_API_KEY, SOURCE=SOURCE)
        self.logging = logging
        self.required_fields = [PRODUCT_NAME_FIELD, PRODUCT_PRICE_FIELD]
        self.projection = []
        self.products_table = table_id
        self.columns = []

    def hash_product(self, product: ProductModel, special_attributes=False) -> str:
        """
        Generates a hash of the product for comparison.
        """
        if special_attributes:
            hash_string = ''.join(attr.description for attr in product.extra_attributes if attr.name.endswith('*'))
        else:
            hash_string = f"{product.external_id}{product.price}{sorted(product.product_properties)}{product.name}{product.description}"

        return hashlib.sha256(hash_string.encode()).hexdigest()

    @cached(ttl=30, key_builder=custom_key_builder)
    async def get_products(self) -> List[ProductModel]:
        """
        Fetches all products from the table.
        """
        records = await self.get_table_records(self.products_table, self.required_fields, self.projection)
        return [parse_product_data(record) for record in records]

    async def get_products_v2(self, offset: int, limit: int) -> List[ProductModel]:
        """
        Fetches paginated products.
        """
        response = await self.get_table_records_v2(
            table_name=self.products_table,
            required_fields=self.required_fields,
            projection=self.projection,
            offset=offset,
            limit=limit,
        )
        return [await parse_product_data(record) for record in response['list']]

    @cached(ttl=180, key_builder=custom_key_builder)
    async def search_products(self, search_string: str, limit: int) -> List[ProductModel]:
        """
        Searches for products with names containing the search string.
        """
        records = await self.get_table_records(
            table_name=self.products_table,
            required_fields=self.required_fields,
            projection=self.projection,
            extra_where=f"({PRODUCT_NAME_FIELD},like,%{search_string}%)",
            limit=limit,
        )
        return [parse_product_data(record) for record in records]

    @cached(ttl=60, key_builder=custom_key_builder)
    async def get_product(self, product_id: str) -> ProductModel:
        """
        Fetches a single product by its ID.
        """
        record = await self.get_table_record(self.products_table, product_id)
        return parse_product_data(record)

    @cached(ttl=60, key_builder=custom_key_builder)
    async def get_product_in_category(self, category_id: Optional[str] = None) -> List[ProductModel]:
        """
        Fetches products within a specific category or all products if no category is specified.
        """
        extra_where = None
        if category_id:
            extra_where = f"({PRODUCT_STOCK_FIELD},gt,0)~and({PRODUCT_CATEGORY_ID_LOOKUP_FIELD},eq,{category_id})"

        records = await self.get_table_records(
            table_name=self.products_table,
            required_fields=self.required_fields,
            projection=self.projection,
            extra_where=extra_where,
        )
        return [parse_product_data(record) for record in records]

    async def update_product(self, product: ProductModel):
        """
        Updates an existing product in the table.
        """
        data = dump_product_data_with_check(data=product, data_check=self.category_manager.categories)
        await self.update_table_record(
            table_name=self.products_table,
            record_id=product.id,
            updated_data=data,
        )
        logger.info(f"Updated product {product.external_id}")

    async def create_product(self, checked_data: dict, product: ProductModel,store_images_at_nocodb: bool=True) -> ProductModel:
        """
        Creates a new product in the table.
        """
        external_id = checked_data.pop("ID")
        metadata = await self.get_table_meta(self.products_table)
        images_column = next(
            column["id"]
            for column in metadata["columns"]
            if column["column_name"] == "Images"
        )

        checked_data[PRODUCT_IMAGES_LOOKUP_FIELD] = [image['title'] for image in checked_data['Images']]

        if store_images_at_nocodb:
            for num, item in enumerate(checked_data['Images']):
                item['url'] = await self.save_image_to_nocodb(
                    source_column_id=self.SOURCE,
                    image_url=item['url'],
                    image_name=item['title'],
                    product_table_name=self.products_table,
                    images_column_id=images_column,
                )

        record = await self.create_table_record(table_name=self.products_table, record=checked_data)
        logger.info(f"Created product {external_id}")
        return record

    async def get_all_products(self) -> List[ProductModel]:
        """
        Fetches all products in paginated portions.
        """
        all_products = []
        portion = 200
        for i in range(10):  # Limit to 10 iterations
            products_portion = await self.get_products_v2(offset=i * portion, limit=portion)
            all_products.extend(products_portion)
            if len(products_portion) < portion:
                break
        return all_products



from typing import List, Optional
from pathlib import Path
import importlib.util

from aiocache import cached
from models.products import ProductModel

# TODO: For test purposes, remove in production
import sys
sys.path.append('/home/latoff/Desktop/MarketBots')

from tgshops_integrations.nocodb_connector.client import NocodbClient
from tgshops_integrations.nocodb_connector.model_mapping import (
    dump_product_data,
    dump_product_data_with_check,
    get_pagination_info,
    ID_FIELD,
    parse_product_data,
    PRODUCT_CATEGORY_ID_LOOKUP_FIELD,
    PRODUCT_NAME_FIELD,
    PRODUCT_PRICE_FIELD,
    PRODUCT_STOCK_FIELD,
    PRODUCT_IMAGES_LOOKUP_FIELD,
    PRODUCT_EXTERNAL_ID,
)
from tgshops_integrations.nocodb_connector.categories_management import CategoryManager
from tgshops_integrations.nocodb_connector.products_management import ProductManager
from loguru import logger

def custom_key_builder(func, *args, **kwargs):
    """
    Key builder function for caching.
    Excludes 'self' by processing args from args[1:].
    """
    args_key_part = "-".join(str(arg) for arg in args[1:])
    kwargs_key_part = "-".join(f"{key}-{value}" for key, value in sorted(kwargs.items()))
    return f"{func.__name__}-{args_key_part}-{kwargs_key_part}"


class Gateway(NocodbClient):

    def __init__(
        self,
        logging: bool = False,
        NOCODB_HOST: Optional[str] = None,
        NOCODB_API_KEY: Optional[str] = None,
        SOURCE: Optional[str] = None,
        filter_buttons: Optional[List[str]] = [],
        config_path: Optional[Path] = None,
        special_attributes: bool = False,
    ):
        super().__init__(NOCODB_HOST=NOCODB_HOST, NOCODB_API_KEY=NOCODB_API_KEY, SOURCE=SOURCE)

        self.logging = logging
        self.required_fields = []
        self.projection = []
        self.special_attributes = special_attributes
        self.filter_buttons = filter_buttons

        if config_path:
            self.load_config_from_path(config_path)

    def load_config_from_path(self, config_path: Path):
        """Loads configuration from the specified path."""
        if config_path.exists():
            spec = importlib.util.spec_from_file_location("config", config_path)
            self.config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.config)
        else:
            raise FileNotFoundError(f"Configuration file not found at {config_path}")

    async def load_data(self, SOURCE: Optional[str] = None):
        """Loads necessary data including tables, categories, and products."""
        self.SOURCE = SOURCE
        self.tables_list = await self.get_all_tables()
        self.products_table = self.tables_list[self.config.NOCODB_PRODUCTS]
        self.category_manager = CategoryManager(
            table_id=self.tables_list[self.config.NOCODB_CATEGORIES],
            NOCODB_HOST=self.NOCODB_HOST,
            NOCODB_API_KEY=self.NOCODB_API_KEY,
            logging=True,
            filter_buttons=self.filter_buttons,
        )
        self.product_manager = ProductManager(
            table_id=self.tables_list[self.config.NOCODB_PRODUCTS],
            NOCODB_HOST=self.NOCODB_HOST,
            NOCODB_API_KEY=self.NOCODB_API_KEY,
            logging=True,
        )

    @cached(ttl=60, key_builder=custom_key_builder)
    async def update_attributes(self, products: List[ProductModel]):
        """Updates attributes for the product table."""
        system_attributes = [PRODUCT_EXTERNAL_ID, PRODUCT_IMAGES_LOOKUP_FIELD]
        attributes = await self.get_table_meta(table_name=self.products_table)
        self.columns = [item['title'].lower() for item in attributes.get('columns', [])]

        # Ensure system attributes exist
        for attribute_name in system_attributes:
            if attribute_name.lower() not in self.columns:
                await self.create_table_column(table_name=self.products_table, name=attribute_name)
                logger.info(f"Created attribute: {attribute_name}")

        # Validate and add extra attributes
        for item in products:
            attributes = await self.get_table_meta(table_name=self.products_table)
            self.columns = [col['title'].lower() for col in attributes.get('columns', [])]

            for attribute in item.extra_attributes:
                if attribute.name.rstrip().lower() not in self.columns:
                    await self.create_table_column(table_name=self.products_table, name=attribute.name.lower())
                    logger.info(f"Created attribute: {attribute.name.lower()}")

    async def update_products(self, external_products: List[ProductModel]):
        """Updates product data by comparing with existing records."""
        await self.update_attributes(products=external_products)
        self.actual_products = await self.product_manager.get_all_products()
        self.ids_mapping = {product.external_id: product.id for product in self.actual_products}
        self.products_meta = {product.external_id: product for product in self.actual_products}

        for product in external_products:
            if (product.external_id in self.ids_mapping and product.external_id !=''):
                product.id = self.ids_mapping[product.external_id]
                current_hash = self.product_manager.hash_product(product, special_attributes=self.special_attributes)
                existing_hash = self.product_manager.hash_product(
                    self.products_meta[product.external_id],
                    special_attributes=self.special_attributes,
                )
                if current_hash != existing_hash:
                    await self.product_manager.update_product(product=product)
            else:
                checked_data = dump_product_data_with_check(
                    data=product,
                    data_check=self.category_manager.categories,
                )
                await self.product_manager.create_product(product=product, checked_data=checked_data)

    async def delete_all_products(self):
        """Deletes all products in the table."""
        items = await self.product_manager.get_products_v2(offset=0, limit=200)
        for item in items:
            await self.product_manager.delete_table_record(self.products_table, item.id)



from setuptools import setup,find_packages

setup(
    name='tgshops_integrations',
    version='4.1',
    packages=find_packages(),
    install_requires=[
        # List your library's dependencies here
        # 'some_package',
    ],
    include_package_data=True,
    description='Library is intended to provide the integration of the external service or CRM system with the TelegramShops/'
                'It allows to configure the relationship between NocoDB list of the products used further to display in the shop/'
                'As a resultss the products can be synchronized and updated uppon the request.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://git.the-devs.com/virtual-shops/shop-system/shop-backend-integrations/integration-library/integration-library',  # Update with your URL
    author='Dimi Latoff',
    author_email='drpozd@gmail.com',
    license='MIT',  # Update with your chosen license
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)

# python3 setup.py sdist bdist_wheel
# twine upload dist/* --verbose